import argparse
import re
from argparse import Namespace

import numpy as np

from ndex.ndpush.boss_resources import BossResParams
from ndex.ndpush.ingest_job import IngestJob
from ndex.ndpush.ingest_large_vol import post_cutout
from ndex.ndpush.parse_log import parse_log


class Cutout:
    def __init__(self, coll, exp, ch, x, y, z):
        self.collection = coll
        self.experiment = exp
        self.channel = ch
        self.x = x
        self.y = y
        self.z = z

        self.log_fname = self.gen_log_fname()

    def cutout_string(self):
        return ('Coll: {}, Exp: {}, Ch: {}, x: {}, y: {}, z: {}'.format(
            self.collection, self.experiment, self.channel, self.x, self.y, self.z))

    def gen_log_fname(self):
        return 'ingest_log_{}_{}_{}_repeat.txt'.format(
            self.collection, self.experiment, self.channel)

    def send_msg(self, msg):
        print(msg)
        with open(self.log_fname, 'a') as f:
            f.write(msg + '\n')


class ImgData():
    def __init__(self, im_array, z_rng):
        self.im_data = im_array
        self.z_rng = z_rng
        self.y = np.shape(im_array)[1]
        self.x = np.shape(im_array)[2]


def parse_cut_line(c_line):
    coll = re.search('Coll: (.+?),', c_line).group(1)
    exp = re.search('Exp: (.+?),', c_line).group(1)
    ch = re.search('Ch: (.+?),', c_line).group(1)
    x = list(map(int, re.search('x: \((.+?)\)', c_line).group(1).split(', ')))
    y = list(map(int, re.search('y: \((.+?)\)', c_line).group(1).split(', ')))
    z = list(map(int, re.search('z: \((.+?)\)', c_line).group(1).split(', ')))

    return coll, exp, ch, x, y, z


def gather_info():
    s = input('Source type (either "local" or "s3"): ')
    if 'local' == s or 's3' == s:
        datasource = s
    else:
        raise TypeError

    if datasource == 's3':
        s = input('s3 bucket name: ')
        s3_bucket_name = s  # not used for 'local' datasource

        s = input('aws profile ("default"): ') or 'default'
        aws_profile = s  # not used for 'local' datasource
    else:
        s3_bucket_name = None
        aws_profile = None

    s = input('BOSS config file path ("neurodata.cfg"): ') or 'neurodata.cfg'
    boss_config_file = s  # location on local system for boss API key

    s = input('data directory (with trailing slash): ')
    # data_directory _with_ trailing slash (doesn't output correct paths on Windows)
    data_directory = s

    # filename without extension (no '.tif')
    # <p:4> indicates the z index of the tif file, with up to 4 leading zeros
    s = input('Filename with  (e.g. "FILENAME<p:4>"): ')
    file_name_pattern = s

    # extension name for images, supported image types are PNG and TIFF
    # extension just needs to match the filename and can be any string (e.g.: ome, tif, png)
    s = input('File format ("tif", "png"): ')
    img_format = s

    # increment of filename numbering (always increment in steps of 1 in the boss, typically will be '1')
    s = input('Z step size ("1"): ') or '1'
    z_step = int(s)

    return Namespace(datasource=datasource,
                     s3_bucket_name=s3_bucket_name,
                     aws_profile=aws_profile,
                     base_filename=file_name_pattern,
                     base_path=data_directory,
                     extension=img_format,
                     z_step=z_step,
                     boss_config_file=boss_config_file,
                     warn_missing_files=True
                     )


def ingest_cuts(cutouts, ingest_job, boss_res_params):
    coll = ingest_job.coll_name
    exp = ingest_job.exp_name
    ch = ingest_job.ch_name

    ingest_job.img_size = [ingest_job.x_extent[1],
                           ingest_job.y_extent[1],
                           ingest_job.z_extent[1]]

    # sort by z, y, x...
    cutouts.sort(key=lambda c: (c.z, c.y, c.x))

    imgdata = None
    for cut in cutouts:
        cut.send_msg(
            'Attempting re-ingest of cutout: {}'.format(cut.cutout_string()))

        # if the data is not in the previous data (same z slice)
        # load the data
        if imgdata is None or imgdata.z_rng != cut.z:
            z_slices = range(cut.z[0], cut.z[1])
            im_array = ingest_job.read_img_stack(z_slices)
            imgdata = ImgData(im_array, cut.z)

        data = imgdata.im_data[:, cut.y[0]:cut.y[1], cut.x[0]:cut.x[1]]
        data = np.asarray(data, order='C')
        ret_val = post_cutout(boss_res_params, ingest_job, cut.x,
                              cut.y, cut.z, data, attempts=2)
        if ret_val == 0:
            cut.send_msg(
                'Successful re-ingest of cutout: {}'.format(cut.cutout_string()))
        else:
            cut.send_msg(
                'Error: re-ingest of cutout failed: {}'.format(cut.cutout_string()))

    cutouts[-1].send_msg('Finished cutouts for collection {}, experiment {}, channel {}'.format(
        coll, exp, ch))


def get_cutouts(repeatfile):
    cutouts = []

    # load the repeat_cutouts file
    with open(repeatfile) as f:
        log_lines = f.readlines()

    # for each line in the repeat cutouts, extract the coll, experiment, channel, x, y, and z
    for line in log_lines:
        coll, exp, ch, x, y, z = parse_cut_line(line)
        cutouts.append(Cutout(coll, exp, ch, x, y, z))

    return cutouts


def iterate_posting_cutouts(cutouts):
    # separate the cutouts into groupings of shared collections/experiments/channels
    collections = set([cu.collection for cu in cutouts])
    for coll in collections:
        cus_coll = [cu for cu in cutouts if cu.collection == coll]
        experiments = set([cu.experiment for cu in cus_coll])
        for exp in experiments:
            cus_exp = [cu for cu in cus_coll if cu.experiment == exp]
            channels = set([cu.channel for cu in cus_exp])
            for ch in channels:
                cus_ch = [cu for cu in cus_exp if cu.channel == ch]
                if len(cus_ch) > 0:
                    # posts data for cutouts that share a common coll, exp, and ch
                    msg = 'Repeating cutouts for collection {}, experiment {}, channel {}'.format(
                        coll, exp, ch)
                    cus_ch[-1].send_msg(msg)

                    args = gather_info()
                    args.collection = coll
                    args.experiment = exp
                    args.channel = ch
                    args.get_extents = True

                    ingest_job = IngestJob(args)
                    # we get these things from the resources that already exist on the boss:
                    boss_res_params = BossResParams(ingest_job)
                    boss_res_params.get_resources(get_only=True)

                    ingest_cuts(cus_ch, ingest_job, boss_res_params)


def main():
    parser = argparse.ArgumentParser(
        description='Search log file for errors and post the data to the boss')
    parser.add_argument('--logfile', type=str,
                        default=None, help='log file to parse')
    parser.add_argument('--repeatfile', type=str,
                        default='repeat_cutouts.txt', help='log file to parse')
    args = parser.parse_args()

    if args.logfile is not None:
        args.repeatfile = parse_log(args.logfile, args.repeatfile)

    cutouts = get_cutouts(args.repeatfile)

    iterate_posting_cutouts(cutouts)

    print('Finished all failed cutouts, check logs for errors')


if __name__ == '__main__':
    main()



'''
Command line program for ingest to boss
Parses arguments and manages the ingest process
'''

import argparse
import platform
import sys
import time
from collections import defaultdict
from datetime import datetime
from functools import partial
from multiprocessing.dummy import Pool as ThreadPool

import numpy as np
from PIL import Image

from ndex.ndpush.boss_resources import BossResParams
from ndex.ndpush.ingest_job import IngestJob

Image.MAX_IMAGE_PIXELS = None


def read_channel_names(channels_path):
    try:
        channels = []
        with open(channels_path, 'r') as f:
            for line in f:
                channels.append(line.strip('\n'))
        return channels
    except FileNotFoundError:
        raise FileNotFoundError


def post_cutout(boss_res_params, ingest_job, x_rng, y_rng, z_rng, data, attempts=5):
    ch = ingest_job.ch_name
    cutout_msg = 'Coll: {}, Exp: {}, Ch: {}, x: {}, y: {}, z: {}'.format(
        ingest_job.coll_name, ingest_job.exp_name, ch, x_rng, y_rng, z_rng)
    # POST cutout
    for attempt in range(attempts):
        try:
            start_time = time.time()
            boss_res_params.rmt.create_cutout(boss_res_params.ch_resource, ingest_job.res,
                                              x_rng, y_rng, z_rng, data)
            end_time = time.time()
            post_time = end_time - start_time
            msg = '{} POST succeeded in {:.2f} sec. {}'.format(
                get_formatted_datetime(), post_time, cutout_msg)
            ingest_job.send_msg(msg)
        except Exception as e:
            # attempt failed
            ingest_job.send_msg(str(e))
            if attempt != attempts - 1:
                time.sleep(2**(attempt + 1))
        else:
            break
    else:
        # we failed all the attempts - deal with the consequences.
        msg = '{} Error: data upload failed after multiple attempts, skipping. {}'.format(
            get_formatted_datetime(), cutout_msg)
        ingest_job.send_msg(msg, send_slack=True)
        ingest_job.num_POST_failures += 1
        return 1
    return 0


def download_boss_slice(boss_res_params, ingest_job, z_slice, attempts=3):
    im_array_boss = np.zeros([1, ingest_job.img_size[1], ingest_job.img_size[0]],
                             dtype=ingest_job.datatype)

    rmt = boss_res_params.rmt
    stride = 2048

    x_buckets = get_supercube_lims(ingest_job.x_extent, stride=stride)
    y_buckets = get_supercube_lims(ingest_job.y_extent, stride=stride)
    for _, y_slices in y_buckets.items():
        y_rng = [y_slices[0],
                 y_slices[-1]+1]
        for _, x_slices in x_buckets.items():
            x_rng = [x_slices[0],
                     x_slices[-1]+1]
            for attempt in range(attempts):
                try:
                    im_array_boss[0, y_rng[0]-ingest_job.y_extent[0]:y_rng[1]-ingest_job.y_extent[0],
                                  x_rng[0]-ingest_job.x_extent[0]:x_rng[-1]-ingest_job.x_extent[0]] = rmt.get_cutout(
                        boss_res_params.ch_resource,
                        ingest_job.res,
                        x_rng, y_rng, [z_slice, z_slice + 1]
                    )
                except Exception as e:
                    # attempt failed
                    ingest_job.send_msg(str(e))
                    if attempt != attempts - 1:
                        time.sleep(2**(attempt + 1))
                else:
                    break
            else:
                # we failed all the attempts - deal with the consequences.
                msg = '{} Error: download cutout failed after multiple attempts'.format(
                    get_formatted_datetime())
                ingest_job.send_msg(msg, send_slack=True)
    return im_array_boss


def assert_equal(boss_res_params, ingest_job, z_rng):
    ingest_job.send_msg('Checking to make sure data has been POSTed correctly')

    # choose a rand slice:
    rand_slice = np.random.randint(z_rng[0], z_rng[1])

    # load data from Boss
    msg = '{} Getting random z slice from BOSS for comparison'.format(
        get_formatted_datetime())
    ingest_job.send_msg(msg)

    im_array_boss = download_boss_slice(
        boss_res_params, ingest_job, rand_slice + ingest_job.offsets[2])

    msg = '{} Z slice from BOSS downloaded'.format(
        get_formatted_datetime())
    ingest_job.send_msg(msg)

    # load source data (one z slice)
    im_array_local = ingest_job.read_img_stack([rand_slice])

    # assert that cutout from the boss is the same as what was sent
    if np.array_equal(im_array_boss, im_array_local):
        ingest_job.send_msg('Test slice {} in Boss matches file {}'.format(
            rand_slice, ingest_job.get_img_fname(rand_slice)))
        return True
    else:
        ingest_job.send_msg('Test slice {} in Boss does *NOT* match file {}'.format(
            rand_slice, ingest_job.get_img_fname(rand_slice)), send_slack=True)
        return False


def get_supercube_lims(rng, stride=16):
    # stride = height of super cuboid

    first = rng[0]    # inclusive
    last = rng[1]     # exclusive

    buckets = defaultdict(list)
    for z in range(first, last):
        buckets[(z // stride)].append(z)

    return buckets


def get_formatted_datetime():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def ingest_block(x_slice_key, x_buckets, boss_res_params, ingest_job, y_rng, z_rng, im_array):
    # created for multithreading
    x_slices = x_buckets[x_slice_key]

    x_rng = [x_slices[0], x_slices[-1] + 1]

    data = im_array[:, y_rng[0]-ingest_job.y_extent[0]:y_rng[1]-ingest_job.y_extent[0],
                    x_rng[0]-ingest_job.x_extent[0]:x_rng[1]-ingest_job.x_extent[0]]
    data = np.asarray(data, order='C')

    if np.sum(data) == 0:
        ingest_job.send_msg('{} Block empty for Collection: {}, Experiment: {}, Channel: {} x/y/z: {}/{}/{}, skipping'.format(
            get_formatted_datetime(),
            ingest_job.coll_name, ingest_job.exp_name, ingest_job.ch_name, x_rng, y_rng, z_rng))
        return

    # POST each block to the BOSS
    post_cutout(boss_res_params, ingest_job,
                x_rng, y_rng, z_rng, data, attempts=3)


def per_channel_ingest(args, channel, threads=8):
    args.channel = channel
    ingest_job = IngestJob(args)

    # extract img_size and datatype to check inputs (by actually reading the data)
    # this can take a while, as we actually load in the first image slice,
    # so we should store this first slice so we don't have to load it again when we later read the entire chunk in z
    # we don't do this for render data source because we get the image size and attributes from the render metadata and the # of bits aren't in the metadata or render
    if ingest_job.datasource != 'render':
        im_width, im_height, im_datatype = ingest_job.get_img_info(
            ingest_job.z_range[0])

        # we do this before creating boss resources that could be inaccurate
        try:
            assert ingest_job.img_size[0] == im_width and ingest_job.img_size[
                1] == im_height and ingest_job.datatype == im_datatype
        except AssertionError:
            ingest_job.send_msg('Mismatch between image file and input parameters. Determined image width: {}, height: {}, datatype: {}'.format(
                im_width, im_height, im_datatype))
            raise ValueError('Image attributes do not match arguments')

    # create or get the boss resources for the data
    get_only = not ingest_job.create_resources
    boss_res_params = BossResParams(ingest_job)
    boss_res_params.get_resources(get_only=get_only)

    # we just create the resources, don't do anything else
    if ingest_job.create_resources:
        ingest_job.send_msg('{} Resources set up. Collection: {}, Experiment: {}, Channel: {}'.format(
            get_formatted_datetime(), ingest_job.coll_name, ingest_job.exp_name, ingest_job.ch_name))
        return 0
    else:
        ingest_job.send_msg('{} Starting ingest for Collection: {}, Experiment: {}, Channel: {}, Z: {z[0]},{z[1]}'.format(
            get_formatted_datetime(), ingest_job.coll_name, ingest_job.exp_name, ingest_job.ch_name, z=ingest_job.z_range))

    # we begin the ingest here:
    stride_x = 1024
    stride_y = 1024
    stride_z = 16
    x_buckets = get_supercube_lims(ingest_job.x_extent, stride_x)
    y_buckets = get_supercube_lims(ingest_job.y_extent, stride_y)
    z_buckets = get_supercube_lims(ingest_job.z_range, stride_z)

    pool = ThreadPool(threads)

    # load images files in stacks of 16 at a time into numpy array
    for _, z_slices in z_buckets.items():
        # read images into numpy array
        im_array = ingest_job.read_img_stack(z_slices)
        z_rng = [z_slices[0] - ingest_job.offsets[2],
                 z_slices[-1] + 1 - ingest_job.offsets[2]]

        # slice into np array blocks
        for _, y_slices in y_buckets.items():
            y_rng = [y_slices[0], y_slices[-1] + 1]

            ingest_block_partial = partial(
                ingest_block, x_buckets=x_buckets, boss_res_params=boss_res_params, ingest_job=ingest_job,
                y_rng=y_rng, z_rng=z_rng, im_array=im_array)
            pool.map(ingest_block_partial, x_buckets.keys())

    # checking data posted correctly for an entire z slice
    assert_equal(boss_res_params, ingest_job, ingest_job.z_range)

    ch_link = (
        'http://ndwt.neurodata.io/channel_detail/{}/{}/{}/').format(ingest_job.coll_name, ingest_job.exp_name, ingest_job.ch_name)

    ingest_job.send_msg('{} Finished z slices {} for Collection: {}, Experiment: {}, Channel: {}\nThere were {} read failures and {} POST failures.\nView properties of channel and start downsample job on ndwebtools: {}'.format(
        get_formatted_datetime(),
        ingest_job.z_range, ingest_job.coll_name, ingest_job.exp_name, ingest_job.ch_name,
        ingest_job.num_READ_failures, ingest_job.num_POST_failures, ch_link), send_slack=True)

    return 0


def main():
    parser = argparse.ArgumentParser(
        description='Copy image z stacks to Boss for a single channel')

    parser.add_argument('--base_path', type=str,
                        help='Directory where image stacks are located (e.g. "/data/images/"')
    parser.add_argument('--base_filename', type=str,
                        help='Base filename with z values specified "ch1_<>" or w/ leading zeros "ch1_<p:4>"')
    parser.add_argument('--extension', type=str, help='Extension (tif(f)/png)')
    parser.add_argument('--datasource', type=str, default='local',
                        help='Location of files, either "local", "s3", or "render"')
    parser.add_argument('--collection', type=str, help='Collection')
    parser.add_argument('--experiment', type=str, help='Experiment')

    parser.add_argument('--channel', type=str, help='Channel')
    parser.add_argument('--channels_list_file', type=str,
                        help='Path to a file with list of channels separated into separate lines')

    parser.add_argument('--voxel_size', type=float,
                        nargs=3, help='Voxel size in x y z')
    parser.add_argument('--voxel_unit', type=str,
                        help='Voxel unit, nanometers/micrometers/millimeters/centimeters')
    parser.add_argument('--datatype', type=str,
                        help='Data type (uint8/uint16/[uint64-annotations])')
    parser.add_argument('--source_channel', type=str,
                        help='Name of reference channel for annotation channels')
    parser.add_argument('--res', type=int, default=0,
                        help='Resolution to copy (default = 0)')
    parser.add_argument('--z_step', type=int, default=1,
                        help='Z step size for input files, default 1 (on Boss, z step is always 1, and z_rng and img_size for z should both assume increments of 1)')
    parser.add_argument('--create_resources', action='store_true',
                        help='Creates the boss resources and exits')

    parser.add_argument('--x_extent', type=int, nargs=2,
                        help='Volume extent in x (width)')
    parser.add_argument('--y_extent', type=int, nargs=2,
                        help='Volume extent in y (height)')
    parser.add_argument('--z_extent', type=int, nargs=2,
                        help='Volume extent in z (slices/images)')
    parser.add_argument('--offset_extents', action='store_true',
                        help='Offset any negative extents to start at zero and store the original values as BOSS metadata')
    parser.add_argument('--forced_offsets', type=int, nargs=3,
                        help='Offset negative extents by a certain value in x/y/z')
    parser.add_argument('--coord_frame_x_extent', type=int, nargs=2,
                        help='Force coordinate frame to be of a partical x extent (must contain data after any offsets)')
    parser.add_argument('--coord_frame_y_extent', type=int, nargs=2,
                        help='Force coordinate frame to be of a partical y extent (must contain data after any offsets)')
    parser.add_argument('--coord_frame_z_extent', type=int, nargs=2,
                        help='Force coordinate frame to be of a partical z extent (must contain data after any offsets)')

    parser.add_argument('--z_range', type=int, nargs=2,
                        help='Z slices to ingest: start (inclusive) end (exclusive)')

    parser.add_argument('--warn_missing_files', action='store_true',
                        help='Warn on missing files instead of failing')

    parser.add_argument('--s3_bucket_name', type=str,
                        help='S3 bucket name')
    parser.add_argument('--aws_profile', type=str, default='default',
                        help='Name of profile in .aws/credentials file (default = default)')

    parser.add_argument('--boss_config_file', type=str, default='neurodata.cfg',
                        help='Path and filename for Boss config (config file w/ server and API Key)')

    parser.add_argument('--slack_token_file', type=str, default='slack_token',
                        help='Path & filename for slack token (key only)')
    parser.add_argument('--slack_usr', type=str,
                        help='User to send slack message to (e.g. USERNAME)')

    parser.add_argument('--render_owner', type=str,
                        help='Name of owner in render')
    parser.add_argument('--render_project', type=str,
                        help='Name of project in render')
    parser.add_argument('--render_stack', type=str,
                        help='Name of stack in render')
    parser.add_argument('--render_channel', type=str,
                        help='Name of channel in render (some stacks don''t have channels so this would not be entered)')
    parser.add_argument('--render_baseURL', type=str,
                        help='Base URL for render instance (https://render-dev-eric.neurodata.io/render-ws/v1/)')
    parser.add_argument('--render_scale', type=float,
                        help='Scale the data imported from render by this factor')
    parser.add_argument('--render_window', type=int, nargs=2,
                        help='Window used on 16bit -> 8 bit data conversion')

    parser.add_argument('--limit_x', type=int, nargs=2,
                        help='Enforced limit in x (down to level of coord frame) to get & post data')
    parser.add_argument('--limit_y', type=int, nargs=2,
                        help='Enforced limit in y (down to level of coord frame) to get & post data')
    parser.add_argument('--limit_z', type=int, nargs=2,
                        help='Enforced limit in z (down to level of coord frame) to get & post data')

    args = parser.parse_args()

    # if not 64 bit python raise an error
    assert platform.architecture()[0] == '64bit'

    # if not python 3.5 or greater, raise an error
    assert sys.version_info >= (3, 5)

    # Iterate through channels if channels path specified
    if args.channels_list_file is not None:
        channels = read_channel_names(args.channels_list_file)
    else:
        channels = [args.channel]

    for channel in channels:
        per_channel_ingest(args, channel)


if __name__ == '__main__':
    main()



import io
import random
import time
from collections import defaultdict
from multiprocessing.dummy import Pool as ThreadPool

import numpy as np
import requests
from PIL import Image

# render web service view
# http://render-dev-eric.neurodata.io/render-ws/view/index.html?

# render stack view
# http://render-dev-eric.neurodata.io/render-ws/view/stacks.html?renderStackOwner=Forrest&renderStackProject=H16_03_005_HSV_HEF1AG65_R2An15dTom&renderStack=ACQTdtomato

# swagger Web Service API
# http://render-dev-eric.neurodata.io/swagger-ui/

# metadata
# https://render-dev-eric.neurodata.io/render-ws/v1/owner/6_ribbon_experiments/project/M321160_Ai139_smallvol/stack/Acquisition_1_PSD95

# image
# https://render-dev-eric.neurodata.io/render-ws/v1/owner/6_ribbon_experiments/project/M321160_Ai139_smallvol/stack/Acquisition_1_PSD95/z/17/box/0,0,4096,4096,0.25/png-image?minIntesnity=0&maxIntensity=5000


class renderResource:
    def __init__(self, owner, project, stack, baseURL, datatype, channel=None, scale=None, limit_x=None, limit_y=None, limit_z=None):
        self.owner = owner
        self.project = project
        self.stack = stack
        self.baseURL = baseURL
        if scale is None:
            self.scale = 1
        else:
            self.scale = scale

        self.datatype = datatype

        # self.level = math.log(1 / scale, 2)
        self.session = requests.Session()

        self.limit_x = limit_x
        self.limit_y = limit_y
        self.limit_z = limit_z

        self.set_metadata()

        if channel:
            assert channel in self.channel_names

        # defaults to None
        self.channel = channel

    def __str__(self):
        from pprint import pformat
        return "<" + type(self).__name__ + "> " + pformat(vars(self), indent=4, width=1)

    def get_render_img(self, z, window=None, threads=1, tile_size=8192):
        # this requests the entire slice and returns the data, scaled if necessary

        # we'll break apart our request into a series of tiles
        # these will extend past the extent of the underlying data
        stride = round(tile_size / self.scale)  # 8K
        x_buckets = get_supercubes(
            self.x_rng_unscaled, stride=stride)
        y_buckets = get_supercubes(
            self.y_rng_unscaled, stride=stride)

        # assembling the args for each of our separate requests
        # requests are set at the unscaled full size resolution
        args = []
        for _, x in x_buckets.items():
            for _, y in y_buckets.items():
                args.append(
                    (z, x[0], y[0], min(stride, x[-1] - x[0] + 1),
                     min(stride, y[-1] - y[0] + 1), window))

        # firing off the requests
        pool = ThreadPool(threads)
        data_array = pool.starmap(self.get_render_tile, args)

        # initialize to the size of the return data (scaled if necessary)
        im_array = np.zeros([tile_size * len(y_buckets),
                             tile_size * len(x_buckets)], dtype=self.datatype)

        # assembling the data
        for idx, data in enumerate(data_array):
            _, x, y, x_width, y_width, _ = args[idx]
            # have to scale the box to fit the data inside
            x_s, y_s = [
                round(a * self.scale) for a in [x, y]]
            y_width, x_width = data.shape
            im_array[y_s - self.y_rng[0]:y_s - self.y_rng[0] + y_width,
                     x_s - self.x_rng[0]:x_s - self.x_rng[0] + x_width] = data

        # we finally clip the data to the bounds of the scaled data (while dealing with offsets)
        im_array = im_array[0:self.y_rng[1] - self.y_rng[0],
                            0:self.x_rng[1] - self.x_rng[0]]
        return im_array

    def set_metadata(self):
        # even if you have a channel the metadata is located at the stack level
        metaURL = '{}owner/{}/project/{}/stack/{}'.format(
            self.baseURL, self.owner, self.project, self.stack)
        r = self.session.get(metaURL, timeout=10)
        if r.status_code != 200:
            raise ConnectionError(
                'Metadata not fetched, error {}'.format(r.reason))
        resp = r.json()
        stats = resp['stats']

        x_start = round(stats['stackBounds']['minX'])
        x_stop = round(stats['stackBounds']['maxX'])
        y_start = round(stats['stackBounds']['minY'])
        y_stop = round(stats['stackBounds']['maxY'])
        z_start = round(stats['stackBounds']['minZ'])
        z_stop = round(stats['stackBounds']['maxZ'])

        self.x_rng_unscaled = [x_start, x_stop]
        self.y_rng_unscaled = [y_start, y_stop]

        self.z_rng = [z_start, z_stop]

        self.validate_xyz_limits()
        self.apply_limits()

        self.x_rng = [round(a * self.scale) for a in self.x_rng_unscaled]
        self.y_rng = [round(a * self.scale) for a in self.y_rng_unscaled]

        self.tile_width = round(stats['maxTileWidth'])
        self.tile_height = round(stats['maxTileHeight'])

        channels = stats['channelNames']
        self.channel_names = channels

    def apply_limits(self):
        if self.limit_x is not None:
            self.x_rng_unscaled = self.limit_x
        if self.limit_y is not None:
            self.y_rng_unscaled = self.limit_y
        if self.limit_z is not None:
            self.z_rng = self.limit_z

    def validate_xyz_limits(self):
        validate_limit(self.x_rng_unscaled, self.limit_x)
        validate_limit(self.y_rng_unscaled, self.limit_y)
        validate_limit(self.z_rng, self.limit_z)

    def gen_render_url(self, z, x, y, x_width, y_width, window=None):
        # using the box cutout

        # this gives you the data at certain levels of downsampling (0 = full res, 1 = half, etc)
        # row and column are the number of rows/colums in the data with the specified width/height
        # GET /v1/owner/{owner}/project/{project}/stack/{stack}/largeDataTileSource/{width}/{height}/{level}/{z}/{row}/{column}.png

        # GET /v1/owner/{owner}/project/{project}/stack/{stack}/z/{z}/box/{x},{y},{width},{height},{scale}/png-image
        # GET /v1/owner/{owner}/project/{project}/stack/{stack}/z/{z}/box/{x},{y},{width},{height},{scale}/png16-image
        if self.datatype == 'uint16':
            png_type = '16'
        else:
            png_type = ''

        img_URL = '{}owner/{}/project/{}/stack/{}/z/{}/box/{},{},{},{},{}/png{}-image'.format(
            self.baseURL, self.owner, self.project, self.stack, z, x, y, x_width, y_width, self.scale, png_type)

        params = []
        if self.channel is not None:
            params.append('channels={}'.format(self.channel))

        if window is not None:
            params.append('minIntensity={}&maxIntensity={}'.format(
                window[0], window[1]))

        if params:
            img_URL += '?' + '&'.join(params)

        return img_URL

    def get_render_tile(self, z, x, y, x_width, y_width, window=None, attempts=6):
        # note that this returns data at scaled resolution, from box coords of unscaled res
        img_URL = self.gen_render_url(z, x, y, x_width, y_width, window=window)

        for attempt in range(attempts):
            try:
                r = self.session.get(img_URL, timeout=60)
                if r.status_code != 200:
                    raise ConnectionError(
                        'Data not fetched with error: {}'.format(r.reason))
            except Exception:
                if attempt != attempts - 1:
                    time.sleep(2**(attempt + 1))
            else:
                break
        else:
            # we failed all the attempts - deal with the consequences.
            raise ConnectionError(
                'Data from URL {} not fetched.  Status code {}, error {}'.format(
                    img_URL, r.status_code, r.reason))

        im_obj = io.BytesIO(r.content)

        if self.datatype == 'uint16':
            return np.array(Image.open(im_obj))
        else:
            return np.array(Image.open(im_obj))[:, :, 0]


def validate_limit(data_rng, limit):
    if limit is not None:
        if limit[0] < data_rng[0] or limit[1] > data_rng[1]:
            raise ValueError


def get_supercubes(rng, stride=512):
    # boss is stored in 512x512x16 so slice the data first and get cutouts in those ranges
    first = rng[0]    # inclusive
    last = rng[1]     # exclusive

    buckets = defaultdict(list)
    for z in range(first, last):
        buckets[(z // stride)].append(z)

    return buckets


def benchmark_get_tile(renderObj, step_size):
    times = []
    for z in range(0, 5):
        x = random.randint(renderObj.x_rng[0], renderObj.x_rng[1] - step_size)
        y = random.randint(renderObj.y_rng[0], renderObj.y_rng[1] - step_size)
        t0 = time.time()
        data = renderObj.get_render_tile(
            z, x, y, step_size, step_size, 1, window=[0, 10000])
        t1 = time.time()
        times.append((t1 - t0) / data.size * 10e5)
    tot_time = np.mean(times)
    print('Step {} took average {:.2f} sec / 10e5 pixels.'.format(step_size, tot_time))


def benchmark_get_img(renderObj, threads, num_runs=10):
    times = []
    for _ in range(0, num_runs):
        z = random.randint(renderObj.z_rng[0], renderObj.z_rng[1])
        t0 = time.time()
        im_array = renderObj.get_render_img(z, scale=1, threads=threads)
        t1 = time.time()
        times.append((t1 - t0))
    tot_time = np.mean(times)
    print('{:2d} threads took {:.2f} sec (avg of {} runs) to extract image of size {}.'.format(
        threads, tot_time, num_runs, im_array.shape))



import json
import time

import blosc
import numpy as np
import requests

BOSS_VERSION = "v1"


class BossMeta:
    def __init__(self, collection, experiment, channel, res=0, iso=False):
        self._collection = collection
        self._experiment = experiment
        self._channel = channel
        self._res = res
        self._iso = iso

    def channel(self):
        return self._channel

    def experiment(self):
        return self._experiment

    def collection(self):
        return self._collection

    def res(self):
        return self._res

    def iso(self):
        return self._iso


class BossRemote:
    def __init__(self, boss_url, token, meta):
        self.boss_url = boss_url
        if self.boss_url[-1] != '/':
            self.boss_url += '/'
        self.token = token

        # BossMeta contains col, exp, chn info
        self.meta = meta

        self.session = requests.Session()
        self.session.headers = {'Authorization': 'Token {}'.format(self.token)}

        self.boss_coll_metadata = self.get_coll_metadata()
        self.boss_exp_metadata = self.get_exp_metadata()
        self.boss_ch_metadata = self.get_channel_metdata()
        self.downsample_status = self.get_downsample_status()

    def __str__(self):
        string = 'Collection: {}, Experiment: {}, Channel: {}\n'.format(
            self.meta.collection(), self.meta.experiment(), self.meta.channel())

        indent_size = 2

        metadata = self.boss_exp_metadata
        metadata_str = json.dumps(metadata, indent=indent_size)
        string += 'Experiment metadata:\n{}\n\n'.format(metadata_str)

        metadata = self.get_coord_frame_metadata(exp_metadata=metadata)
        metadata_str = json.dumps(metadata, indent=indent_size)
        string += 'Coordinate frame metadata:\n{}\n\n'.format(metadata_str)

        metadata = self.boss_ch_metadata
        metadata_str = json.dumps(metadata, indent=indent_size)
        string += 'Channel metadata:\n{}\n\n'.format(metadata_str)

        return string

    def get(self, url, headers={}):
        if url[0] == '/':
            url = url[1:]
        r = self.session.get("{}{}".format(
            self.boss_url, url), headers=headers)
        return r

    def get_coll_metadata(self):
        # https://api.theboss.io/v1/collection/:collection/
        url = "{}/collection/{}".format(
            BOSS_VERSION, self.meta.collection()
        )
        resp = self.get(url, {'Accept': 'application/json'})
        return resp.json()

    def get_exp_metadata(self):
        # https://api.theboss.io/v1/collection/:collection/experiment/:experiment/
        exp_url = "{}/collection/{}/experiment/{}/".format(
            BOSS_VERSION, self.meta.collection(), self.meta.experiment()
        )
        resp = self.get(exp_url, {'Accept': 'application/json'})
        return resp.json()

    def get_channel_metdata(self):
        # https://api.boss.neurodata.io/v1/collection/:collection/experiment/:experiment/channel/:channel/
        ch_url = '{}/collection/{}/experiment/{}/channel/{}/'.format(
            BOSS_VERSION, self.meta.collection(), self.meta.experiment(), self.meta.channel()
        )
        resp = self.get(ch_url, {'Accept': 'application/json'})
        return resp.json()

    def get_coord_frame_metadata(self, exp_metadata=None):
        coord_frame_name = self.get_coord_frame_name(exp_metadata)
        coord_frame_url = "{}/coord/{}".format(BOSS_VERSION, coord_frame_name)
        resp = self.get(coord_frame_url, {'Accept': 'application/json'})
        return resp.json()

    def get_downsample_status(self):
        # https://api.boss.neurodata.io/v1/downsample/kristina15/image/CR1_2ndA
        url = "{}/downsample/{}/{}/{}".format(
            BOSS_VERSION, self.meta.collection(), self.meta.experiment(), self.meta.channel())
        resp = self.get(url, {'Accept': 'application/json'})
        return resp.json()

    def get_coord_frame_name(self, exp_data=None):
        if exp_data is None:
            exp_data = self.boss_exp_metadata
        return exp_data['coord_frame']

    def get_xyz_extents(self):
        coord_frame = self.get_coord_frame_metadata()
        x_rng = [coord_frame['x_start'], coord_frame['x_stop']]
        y_rng = [coord_frame['y_start'], coord_frame['y_stop']]
        z_rng = [coord_frame['z_start'], coord_frame['z_stop']]

        # extents are different in x/y for downsampled data
        x_rng, y_rng = [[round(bnd / 2**self.meta.res()) for bnd in rng]
                        for rng in [x_rng, y_rng]]

        if self.meta.iso():
            # need to convert to list to set new z_rng
            z_rng = [int(z / 2**self.meta.res()) for z in z_rng]

        return x_rng, y_rng, z_rng

    def cutout(self, x_rng, y_rng, z_rng, datatype, attempts=5):
        cutout_url_base = "{}/cutout/{}/{}/{}".format(
            BOSS_VERSION, self.meta.collection(), self.meta.experiment(), self.meta.channel())
        cutout_url = "{}/{}/{}:{}/{}:{}/{}:{}/".format(
            cutout_url_base, self.meta.res(), x_rng[0], x_rng[1], y_rng[0], y_rng[1], z_rng[0], z_rng[1])
        if self.meta.iso():
            cutout_url += '?iso=True'

        for attempt in range(attempts):
            try:
                resp = self.get(cutout_url, {'Accept': 'application/blosc'})
                resp.raise_for_status()
            except Exception:
                if attempt != attempts - 1:
                    time.sleep(2**(attempt + 1))
            else:
                break
        else:
            # we failed all the attempts - deal with the consequences.
            raise ConnectionError(
                'Data from URL {} not fetched.  Status code {}, error {}'.format(
                    cutout_url, resp.status_code, resp.reason))

        raw_data = blosc.decompress(resp.content)
        data = np.fromstring(raw_data, dtype=datatype)

        return np.reshape(data,
                          (z_rng[1] - z_rng[0],
                           y_rng[1] - y_rng[0],
                           x_rng[1] - x_rng[0]),
                          order='C')



'''
Class for each ingest job (per channel)
Stores all the info for each ingest job
'''

import io
import os
import re
import time
from datetime import datetime

import boto3
import numpy as np
import tailer
import tifffile
from PIL import Image
from slacker import Slacker

from ndex.ndpush.render_resource import renderResource


class IngestJob:
    def __init__(self, args_namespace):
        # args is a Namespace (argparse)

        # convert to dict in order to use the get method (returns None if not present in dict)
        args = vars(args_namespace)

        self.datasource = args.get('datasource')

        self.slack_usr = args.get('slack_usr')
        self.s3_bucket_name = args.get('s3_bucket_name')
        self.create_resources = args.get('create_resources')
        self.coll_name = args.get('collection')
        self.exp_name = args.get('experiment')
        self.ch_name = args.get('channel')
        self.datatype = args.get('datatype')
        self.warn_missing_files = args.get('warn_missing_files')
        self.z_range = args.get('z_range')

        self.limit_x = args.get('limit_x')
        self.limit_y = args.get('limit_y')
        self.limit_z = args.get('limit_z')

        self.voxel_size = args.get('voxel_size')
        self.voxel_unit = args.get('voxel_unit')
        self.res = args.get('res')
        self.source_channel = args.get('source_channel')

        if self.source_channel is not None:
            self.boss_datatype = 'uint64'  # force boss datatype to uint64 for annotations
            self.ch_type = 'annotation'
        else:
            # image channels don't have a source
            try:
                assert self.datatype == 'uint8' or self.datatype == 'uint16' or self.datatype is None
            except AssertionError:
                raise ValueError(
                    'image channels in the BOSS only support uint8 or uint16')
            self.boss_datatype = self.datatype
            self.ch_type = 'image'

        if self.datasource == 'render':
            # create a render resource which will populate some of the variables
            render_owner = args.get('render_owner')
            render_project = args.get('render_project')
            render_stack = args.get('render_stack')
            render_channel = args.get('render_channel')
            render_baseURL = args.get('render_baseURL')

            render_scale = args.get('render_scale')
            self.render_window = args.get('render_window')

            # create the render object in order to get the xyz extents
            self.render_obj = renderResource(render_owner, render_project, render_stack, render_baseURL, self.datatype,
                                             channel=render_channel, scale=render_scale, limit_x=self.limit_x, limit_y=self.limit_y, limit_z=self.limit_z)
            self.x_extent = self.render_obj.x_rng
            self.y_extent = self.render_obj.y_rng
            self.z_extent = self.render_obj.z_rng

            self.z_step = 1
            # render resource is set to return PNGs
            self.extension = 'png'
            if self.z_range is None:  # if the user isn't specifying the z range for ingest, we just get the entire extent
                self.z_range = self.z_extent

        # otherwise it's image data
        elif self.datasource == 's3' or self.datasource == 'local':
            self.base_fname = args.get('base_filename')
            self.base_path = args.get('base_path')
            self.extension = args.get('extension')
            self.x_extent = args.get('x_extent')
            self.y_extent = args.get('y_extent')
            self.z_extent = args.get('z_extent')
            self.validate_xyz_limits()
            self.apply_limits()
            self.z_step = args.get('z_step')

        # initialize offset to zero (x,y,z)
        self.offsets = [0, 0, 0]
        self.forced_offsets = args.get('forced_offsets')
        # if user specified, we calculate the offset
        if args.get('offset_extents') or self.forced_offsets is not None:
            # calculate and store offsets
            self.offsets = self.calc_offsets()

            # create new extents
            self.offset_extents()

        self.set_img_size()

        self.coord_frame_x_extent = args.get('coord_frame_x_extent')
        self.coord_frame_y_extent = args.get('coord_frame_y_extent')
        self.coord_frame_z_extent = args.get('coord_frame_z_extent')
        self.set_coord_frame_extents()
        self.get_extents = args.get('get_extents')
        self.validate_coord_frames()

        # creating the slack session
        self.slack_obj = self.create_slack_session(
            args.get('slack_token_file'))

        if self.datasource == 's3':
            self.s3_res = self.create_s3_res(
                aws_profile=args.get('aws_profile'))
        else:
            self.s3_res = None

        self.boss_config_file = args.get('boss_config_file')

        self.num_READ_failures = 0
        self.num_POST_failures = 0

        # Document the arguments passed
        self.send_msg('{} Command parameters used: {}'.format(
            get_formatted_datetime(), args))

    def apply_limits(self):
        if self.limit_x is not None:
            self.x_rng_unscaled = self.limit_x
        if self.limit_y is not None:
            self.y_rng_unscaled = self.limit_y
        if self.limit_z is not None:
            self.z_rng = self.limit_z

    def validate_xyz_limits(self):
        validate_limit(self.x_extent, self.limit_x)
        validate_limit(self.y_extent, self.limit_y)
        validate_limit(self.z_extent, self.limit_z)

        # z range is a limit itself - we check we aren't going over our limit with z_range
        validate_limit(self.limit_z, self.z_range)

    def validate_coord_frames(self):
        coord_extents = [self.coord_frame_x_extent,
                         self.coord_frame_y_extent,
                         self.coord_frame_z_extent]
        extents = [self.x_extent, self.y_extent, self.z_extent]

        if (None in extents or None in coord_extents) and (not self.get_extents):
            raise ValueError

        if not self.get_extents:
            for coord, ext in zip(coord_extents, extents):
                if coord[0] > ext[0] or coord[1] < ext[1]:
                    raise ValueError

    def create_slack_session(self, slack_token_file):
        if self.slack_usr is None:
            self.send_msg(
                'Slack user not specified, Slack messages will not be sent')
            return None

        if slack_token_file:
            # generate token here: https://api.slack.com/custom-integrations/legacy-tokens, put in file in same directory -> "slack_token"
            try:
                with open(slack_token_file, 'r') as s:
                    token = s.readline().split("\n")
                    token = token[0]
            except FileNotFoundError:
                self.send_msg('Slack token file not found: {}, create slack_token file for sending Slack messages'.format(
                    slack_token_file))
                return None
        else:
            # attempt to load from environment variable
            try:
                token = os.environ['SLACK_TOKEN']
            except Exception as e:
                print(e)
                return None

        return Slacker(token)

    def create_s3_res(self, aws_profile='default'):
        # initiating the S3 resource:
        if self.datasource == 's3':
            if self.s3_bucket_name is None:
                raise ValueError(
                    's3 bucket not defined but s3 datasource chosen')
            try:
                s3_session = boto3.session.Session(profile_name=aws_profile)
                s3_res = s3_session.resource('s3')
            except ValueError:
                raise ValueError('AWS credentials not set up?')
        else:
            if self.s3_bucket_name is not None:
                self.send_msg('s3 bucket name input but source is local')
            s3_res = None
        return s3_res

    def get_log_fname(self):
        return '_'.join(('ingest_log', self.coll_name, self.exp_name, self.ch_name)) + '.txt'

    def send_msg(self, msg, send_slack=False):
        logfile = self.get_log_fname()

        print(msg)
        with open(logfile, 'a') as f:
            f.write(msg + '\n')
        if send_slack and self.slack_obj is not None:
            self.slack_obj.chat.post_message(
                '@' + self.slack_usr, msg, username='local_ingest.py')
            content = tailer.tail(open(logfile), 10)
            self.slack_obj.files.upload(content='\n'.join(
                content), channels='@' + self.slack_usr, title=get_formatted_datetime() + '_tail_of_log')

    def calc_offsets(self):
        if self.forced_offsets is not None:
            return self.forced_offsets

        offsets = []
        for extent in [self.x_extent, self.y_extent, self.z_extent]:
            offset = 0
            if extent[0] < 0:
                offset = abs(extent[0])
            offsets.append(offset)
        return offsets

    def offset_extents(self):
        self.x_extent, self.y_extent, self.z_extent = [
            [ext[0] + off, ext[1] + off] for ext, off in zip([self.x_extent, self.y_extent, self.z_extent], self.offsets)]

    def set_coord_frame_extents(self):
        if self.coord_frame_x_extent is None:
            self.coord_frame_x_extent = self.x_extent
        if self.coord_frame_y_extent is None:
            self.coord_frame_y_extent = self.y_extent
        if self.coord_frame_z_extent is None:
            self.coord_frame_z_extent = self.z_extent

    def set_img_size(self):
        # validation that xyz extents are not negative
        # they could be None if just trying to get the Boss resource (not set it)
        if all([a is not None for a in [self.x_extent, self.y_extent, self.z_extent]]):
            try:
                assert all(
                    [a >= 0 for a in [self.x_extent[0], self.y_extent[0], self.z_extent[0]]])
            except AssertionError:
                raise ValueError('Extents must be positive for the BOSS')

            # helper variable
            self.img_size = [self.x_extent[1] - self.x_extent[0],
                             self.y_extent[1] - self.y_extent[0],
                             self.z_extent[1] - self.z_extent[0]]
        else:
            self.img_size = None

    def get_img_info(self, z_slice):
        img = self.load_img(z_slice)

        width = img.shape[1]
        height = img.shape[0]
        datatype = img.dtype
        return (width, height, datatype)

    def validate_local_img(self, img_fname):
        if not os.path.isfile(img_fname):
            msg = '{} File not found: {}'.format(
                get_formatted_datetime(), img_fname)

            self.send_msg(msg, send_slack=True)
            if self.warn_missing_files:
                return None
            else:
                raise IOError(msg)
        return img_fname

    def load_s3_obj(self, img_fname, attempts=3):
        for attempt in range(attempts):
            try:
                obj = self.s3_res.Object(self.s3_bucket_name, img_fname)
                return io.BytesIO(obj.get()['Body'].read())
            except Exception as err:
                msg = '{} Exception {} occurred when getting image {} from s3'.format(
                    get_formatted_datetime(), err, img_fname)
                if attempt != attempts - 1:
                    time.sleep(2**(attempt + 1))

        self.send_msg(msg, send_slack=True)
        self.num_READ_failures += 1
        if self.warn_missing_files:
            return None
        else:
            raise IOError(msg)

    def load_render_slice(self, z_slice):
        self.send_msg('{} Getting slice {} from render.'.format(
            get_formatted_datetime(), z_slice))
        try:
            return self.render_obj.get_render_img(z_slice, window=self.render_window)
        except Exception as err:
            msg = '{} Exception {} occurred when getting image {} from render with error message {}'.format(
                get_formatted_datetime(), err, z_slice, str(err))
            self.num_READ_failures += 1
            if self.warn_missing_files:
                return None
            else:
                raise IOError(msg)

    def load_img(self, z_slice):
        if self.datasource == 'render':
            # download the slice from render server
            return self.load_render_slice(z_slice)

        # if it's not render datasource, we are working with images in some form
        img_fname = self.get_img_fname(z_slice)
        if self.datasource == 'local':
            # ensure the image exists on filesystem
            im_obj = self.validate_local_img(img_fname)
        elif self.datasource == 's3':
            # download the file from s3
            im_obj = self.load_s3_obj(img_fname)

        # called if datasource is s3 or local
        try:
            _, extension = os.path.splitext(img_fname)
            # if it's PNG we load it with PILLOW using the user specified datatype
            if extension.lower() == '.png':
                im = np.array(Image.open(im_obj), dtype=self.datatype)

            # if it is ome, load with appropriate kwarg
            elif extension.lower() == '.ome':
                im = tifffile.imread(im_obj, is_ome=True)
            # if it's not ome, avoid loading ome metadata
            # bug fix sometimes for .ome.tif files
            else:
                im = tifffile.imread(im_obj, is_ome=False)

            return im

        except OSError:
            msg = '{} Problem opening file: {}'.format(
                get_formatted_datetime(), img_fname)
            self.send_msg(msg, send_slack=True)
            self.num_READ_failures += 1
            if self.warn_missing_files:
                return None
            raise OSError(msg)

        except Exception as err:
            msg = '{} Unknown error {}: {}'.format(
                get_formatted_datetime(), err, img_fname)
            self.send_msg(msg, send_slack=True)
            self.num_READ_failures += 1
            if self.warn_missing_files:
                return None
            raise IOError(msg)

    def get_img_fname(self, z_index):
        if self.datasource == 'render':
            return None

        base_path = self.base_path
        base_fname = self.base_fname

        if z_index >= self.z_range[1]:
            raise IndexError("Z-index out of range")

        matches = re.findall(r'<(p:\d+)?>', base_fname)
        for m in matches:
            if m:
                # There is zero padding
                z_str = str(z_index * self.z_step).zfill(int(m.split(':')[1]))
            else:
                z_str = str(z_index * self.z_step)
            base_fname = base_fname.replace("<{}>".format(m), z_str)

        # replace <ch> in filename with channel.name
        matches = re.findall('<(ch)>', base_fname)
        for m in matches:
            base_fname = base_fname.replace(
                "<{}>".format(m), self.ch_name)

        # replace <ch> in filename with channel.name
        matches = re.findall('<(ch)>', base_path)
        for m in matches:
            base_path = base_path.replace(
                "<{}>".format(m), self.ch_name)

        # prepend root, append extension
        return os.path.join(base_path, "{}.{}".format(base_fname, self.extension))

    def read_img_stack(self, z_slices):
        self.send_msg('{} Reading image data (z range: {}:{})'.format(
            get_formatted_datetime(), z_slices[0], z_slices[-1] + 1))

        start_time = time.time()
        im_array = np.zeros(
            (len(z_slices), self.img_size[1], self.img_size[0]), dtype=self.datatype, order='C')
        for idx, z_slice in enumerate(z_slices):
            img = self.load_img(z_slice)
            if img is None and self.warn_missing_files:
                continue
            im_array[idx, :, :] = img

        # cast the data as uint64 for the BOSS annotations even if the data is something else
        if self.datatype != 'uint64' and self.boss_datatype == 'uint64':
            im_array = im_array.astype('uint64')

        end_time = time.time()
        read_time = end_time - start_time
        self.send_msg('{} Finished reading image data (z range: {}:{}) in {:.2f} sec'.format(
            get_formatted_datetime(), z_slices[0], z_slices[-1] + 1, read_time))
        return im_array


def get_formatted_datetime():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def validate_limit(data_rng, limit):
    if data_rng is not None and limit is not None:
        if limit[0] < data_rng[0] or limit[1] > data_rng[1]:
            raise ValueError



'''
Command line program to download a region of data from the BOSS as individual TIFF slices
'''

import argparse
import configparser
import json
import math
import os
import sys
import time
from collections import defaultdict
from functools import partial
from multiprocessing.dummy import Pool as ThreadPool
from pathlib import Path

import blosc
import numpy as np
import requests
import tifffile as tiff
from tqdm import tqdm

from ndex.ndpull.boss_resources import *

# download blocks of size 2k by 2k by 16 (xyz), should be multiples of 512 x 512 x 16
CHUNK_SIZE = (2048, 2048, 16)


def get_cube_lims(rng, stride=16):
    # stride = height of super cuboid

    first = rng[0]    # inclusive
    last = rng[1]     # exclusive

    buckets = defaultdict(list)
    for z in range(first, last):
        buckets[(z // stride)].append(z)

    return buckets


def collect_input_args(collection, experiment, channel, config_file=None, token=None, url='https://api.boss.neurodata.io', x=None, y=None, z=None, res=0, outdir='./', full_extent=False, print_metadata=False, iso=False):
    result = argparse.Namespace(
        collection=collection,
        experiment=experiment,
        channel=channel,
        config_file=config_file,
        token=token,
        url=url,
        x=x, y=y, z=z, res=res, outdir=outdir,
        full_extent=full_extent,
        print_metadata=print_metadata,
        iso=iso,
    )
    return result


def collect_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config_file', type=str,
                        help='User config file for BOSS')
    parser.add_argument(
        '--token', type=str, help='User token for the boss (not used if config file specified)')
    parser.add_argument(
        '--url', default='https://api.boss.neurodata.io', help='URL to boss endpoint (not used if config file specified)')

    parser.add_argument('--collection', type=str, help='Collection')
    parser.add_argument('--experiment', type=str, help='Experiment')
    parser.add_argument('--channel', type=str, help='Channel')

    # parser.add_argument('--download_limit', type=int, default=500,
    #                     help='limit total download size to this value in GB')

    parser.add_argument('--x', nargs=2, type=int, help='X range for stack')
    parser.add_argument('--y', nargs=2, type=int, help='Y range for stack')
    parser.add_argument('--z', nargs=2, type=int, help='Z range for stack')
    parser.add_argument('--res', default=0, type=int, help='Stack resolution')

    parser.add_argument('--outdir', type=str, default="./",
                        help='Path to output directory.')

    parser.add_argument('--full_extent', action='store_true',
                        help='Use the full extent of the data on the BOSS')

    parser.add_argument('--print_metadata', action='store_true',
                        help='Prints the metadata on the collection/experiment/channel and quits')
    parser.add_argument('--threads', default=4, type=int,
                        help='Number of threads for downloading data.')
    parser.add_argument('--iso', action='store_true',
                        help='Returns iso data (for downsampling in z)')

    parser.add_argument('--stack_filename', type=str,
                        help='If specified, tiffs are merged into a single tif stack file, at the outdir specified')

    parser.add_argument('--force_datatype', type=str,
                        help='downloaded data will be cast into this datatype (uint8/uint16/uint32)')

    return parser.parse_args()


def get_boss_config(boss_config_file=None):
    config = configparser.ConfigParser()

    if boss_config_file:
        config.read(boss_config_file)
        token = config['Default']['token']
        protocol = config['Default']['protocol']
        host = config['Default']['host']
    else:
        token = os.environ['BOSS_TOKEN']
        protocol = 'https'
        host = 'api.boss.neurodata.io'

    boss_url = ''.join((protocol, '://', host))
    return token, boss_url


def download_slices(result, rmt, threads=4):

    # get the datatype
    ch_meta = rmt.boss_ch_metadata
    datatype = ch_meta['datatype']

    z_buckets = get_cube_lims(result.z, stride=CHUNK_SIZE[2])
    for _, z_slices in tqdm(z_buckets.items()):
        z_rng = [z_slices[0], z_slices[-1] + 1]

        # re-initialize on every slice of z
        # zyx ordered
        data_slices = np.zeros((z_rng[1] - z_rng[0],
                                result.y[1] - result.y[0],
                                result.x[1] - result.x[0]),
                               dtype=datatype)

        for y in range(result.y[0], result.y[1], CHUNK_SIZE[1]):
            y_rng = [y, result.y[1] if (
                y + CHUNK_SIZE[1]) > result.y[1] else (y + CHUNK_SIZE[1])]

            x_rngs = []
            for x in range(result.x[0], result.x[1], CHUNK_SIZE[0]):
                x_rngs.append([x, result.x[1] if (
                    x + CHUNK_SIZE[0]) > result.x[1] else (x + CHUNK_SIZE[0])])

            cutout_partial = partial(
                rmt.cutout, y_rng=y_rng, z_rng=z_rng, datatype=datatype)
            with ThreadPool(threads) as pool:
                data_list = pool.map(cutout_partial, x_rngs)

            for data, x_rng in zip(data_list, x_rngs):
                # insert into numpy array
                data_slices[:,
                            y_rng[0] - result.y[0]:y_rng[1] - result.y[0],
                            x_rng[0] - result.x[0]:x_rng[1] - result.x[0]] = data

        save_to_tiffs(data_slices, rmt.meta, result,
                      z_rng, result.force_datatype)


def gen_tif_fname(meta, result, zslice, digits):
    file_format = '{}_{}_{}_x{x[0]}-{x[1]}_y{y[0]}-{y[1]}_z{z:0{dig}d}.tif'
    fname = file_format.format(
        meta.collection(), meta.experiment(), meta.channel(),
        x=result.x, y=result.y, z=zslice, dig=digits)
    return fname


def save_to_tiffs(data_slices, meta, result, z_rng, force_datatype=None):
    # save the numpy array as a tiff file

    # path for saving slices
    cutout_path = Path(result.outdir)
    cutout_path.mkdir(parents=True, exist_ok=True)

    digits = int(math.log10(result.z[1])) + 1

    for zslice in range(z_rng[0], z_rng[1]):
        fname = gen_tif_fname(meta, result, zslice, digits)

        data = data_slices[zslice - z_rng[0], :, :]

        if force_datatype:
            data = data.astype(force_datatype)

        tiff.imsave(str(cutout_path / fname), data,
                    metadata={'DocumentName': fname}, compress=6)


def save_to_stack(meta, result):
    # save the extracted slices to a single tiff stack

    cutout_path = Path(result.outdir)

    stack_fname = Path(cutout_path, result.stack_filename)

    digits = int(math.log10(result.z[1])) + 1

    try:
        stack_fname.unlink()
    except OSError:
        pass
    for zslice in range(result.z[0], result.z[-1]):
        img_fname = gen_tif_fname(meta, result, zslice, digits)

        I = tiff.imread(str(cutout_path / img_fname))
        tiff.imsave(str(stack_fname), I, append=True)
        Path(cutout_path / img_fname).unlink()

# to add:
# tracking of amount of data requests (with default limit, so it can stop if it passes a threshold)


def validate_args(result):
    # get tokens from config file if set as an option
    result.token, result.url = get_boss_config(result.config_file)
    if result.token is None:
        error_msg = 'Need token or config file'
        print(error_msg)
        raise ValueError(error_msg)

    meta = BossMeta(result.collection, result.experiment,
                    result.channel, result.res, result.iso)
    rmt = BossRemote(result.url, result.token, meta)

    if result.print_metadata:
        print(rmt)
        sys.exit()

    if meta.res() > 0:
        if meta.res() >= rmt.boss_exp_metadata['num_hierarchy_levels']:
            raise ValueError('Res argument too high for experiment')

        if rmt.downsample_status['status'] != 'DOWNSAMPLED':
            raise ValueError('Experiment not downsampled')

    # list of full extent in xyz
    full_range = rmt.get_xyz_extents()

    if result.full_extent:
        if any([a is not None for a in [result.x, result.y, result.z]]):
            error_msg = 'full extent not compatible with specified x, y, or z ranges'
            print(error_msg)
            raise ValueError(error_msg)
        result.x, result.y, result.z = full_range
    else:
        if any([a is None for a in [result.x, result.y, result.z]]):
            error_msg = 'Need extents for x, y, and z for download'
            print(error_msg)
            raise ValueError(error_msg)
        if not all([a[0] >= b[0] and a[1] <= b[1]
                    for a, b in zip([result.x, result.y, result.z], full_range)]):
            error_msg = 'Some values out of range for experiment: {} (x, y, z)'.format(
                full_range)
            print(error_msg)
            raise ValueError(error_msg)

    return result, rmt


def main():
    args = collect_args()
    result, rmt = validate_args(args)

    print('Starting download')
    download_slices(result, rmt, threads=args.threads)
    print('Download complete')

    if result.stack_filename:
        save_to_stack(rmt.meta, result)


if __name__ == '__main__':
    main()



import argparse
import os

from tqdm import tqdm


def get_nonexistant_path(fname_path):
    """
    Get the path to a filename which does not exist by incrementing path.

    Examples
    --------
    >>> get_nonexistant_path('/etc/issue')
    '/etc/issue-1'
    >>> get_nonexistant_path('whatever/1337bla.py')
    'whatever/1337bla.py'
    """
    if not os.path.exists(fname_path):
        return fname_path
    filename, file_extension = os.path.splitext(fname_path)
    i = 1
    new_fname = "{}-{}{}".format(filename, i, file_extension)
    while os.path.exists(new_fname):
        i += 1
        new_fname = "{}-{}{}".format(filename, i, file_extension)
    return new_fname


def search(lines, text):
    results = []
    for line in lines:
        if text in line:
            results.append(line)
    return results


def parse_log(logfile, outfile):
    # parse the log file to generate the repeat_cutouts file

    if os.path.isfile(outfile):
        outfile = get_nonexistant_path(outfile)

    # this just creates and empties the outfile if it already exists
    with open(outfile, 'w'):
        pass

    # read the entire thing into memory because we have to search the thing multiple times
    with open(logfile) as f:
        log_lines = f.readlines()

    error_lines = search(log_lines, ', skipping')
    success_lines = search(log_lines, 'POST succeeded')

    for line in tqdm(error_lines):
        line_split = line.split(' Coll: ')
        cutout = 'Coll: {}'.format(line_split[1]).strip('\n')
        cutout_lines = search(success_lines, cutout)
        if not cutout_lines:
            with open(outfile, 'a') as fo:
                fo.write(cutout + '\n')

    return outfile


def main():
    parser = argparse.ArgumentParser(description='Search log file for errors')
    parser.add_argument('--logfile', type=str,
                        default='log.txt', help='log file to parse')
    parser.add_argument('--outfile', type=str,
                        default='repeat_cutouts.txt', help='log file to parse')
    args = parser.parse_args()

    print('Parsing the log file...')
    parse_log(args.logfile, args.outfile)
    print('Parsing log file complete.')


if __name__ == '__main__':
    main()



'''
Class for interacting with the BOSS
Stores BOSS resources
Associated with an ingest job
'''

import math
import os

from requests import HTTPError

from intern.remote.boss import BossRemote
from intern.resource.boss.resource import *
from intern.service.boss.httperrorlist import HTTPErrorList


class BossResParams:
    def __init__(self, ingest_job):
        self.ingest_job = ingest_job

        self.coord_frame_name = '_'.join(
            (ingest_job.coll_name, ingest_job.exp_name))

        self.rmt = self.setup_remote()

    def get_resources(self, get_only=True):
        self.coll_resource = self.setup_boss_collection(get_only=get_only)

        self.coord_frame_resource = self.setup_boss_coord_frame(
            get_only=get_only)

        self.exp_resource = self.setup_boss_experiment(get_only=get_only)

        self.ch_resource = self.setup_boss_channel(get_only=get_only)

    def setup_remote(self):
        if self.ingest_job.boss_config_file:
            return BossRemote(self.ingest_job.boss_config_file)
        else:
            # try to load from environment variable
            token = os.environ['BOSS_TOKEN']
            protocol = 'https'
            host = 'api.boss.neurodata.io'
            config_dict = {'token': token, 'protocol': protocol, 'host': host}
            return BossRemote(config_dict)

    def get_boss_project(self, proj_setup, get_only):
        try:
            proj_actual = self.rmt.get_project(proj_setup)
        except HTTPError as e:
            if get_only is False:
                try:
                    proj_actual = self.rmt.create_project(proj_setup)
                except Exception as e:
                    print(type(e), e)
                    raise(e)
            else:
                print(type(e), e)
                raise e
        except Exception as e:
            print(type(e), e)
            raise(e)
        return proj_actual

    def setup_boss_collection(self, get_only=True):
        coll_setup = CollectionResource(self.ingest_job.coll_name)
        return self.get_boss_project(coll_setup, get_only)

    def setup_boss_coord_frame(self, get_only=True):
        # if we don't know the coordinate frame parameters, get the one with the same name
        if get_only:
            coord_setup = CoordinateFrameResource(self.coord_frame_name)
        else:
            coord_setup = CoordinateFrameResource(self.coord_frame_name, '',
                                                  self.ingest_job.coord_frame_x_extent[0],
                                                  self.ingest_job.coord_frame_x_extent[1],
                                                  self.ingest_job.coord_frame_y_extent[0],
                                                  self.ingest_job.coord_frame_y_extent[1],
                                                  self.ingest_job.coord_frame_z_extent[0],
                                                  self.ingest_job.coord_frame_z_extent[1],
                                                  self.ingest_job.voxel_size[0],
                                                  self.ingest_job.voxel_size[1],
                                                  self.ingest_job.voxel_size[2],
                                                  self.ingest_job.voxel_unit)
        coord_frame_resource = self.get_boss_project(coord_setup, get_only)

        if get_only:
            # matching ingest_job values to coordinate frame values (if they weren't specified, they are now populated)
            self.ingest_job.voxel_size = [coord_frame_resource.x_voxel_size,
                                          coord_frame_resource.y_voxel_size,
                                          coord_frame_resource.z_voxel_size]
            self.ingest_job.voxel_unit = coord_frame_resource.voxel_unit
            if self.ingest_job.get_extents:
                self.ingest_job.coord_frame_x_extent = [coord_frame_resource.x_start,
                                                        coord_frame_resource.x_stop]
                self.ingest_job.coord_frame_y_extent = [coord_frame_resource.y_start,
                                                        coord_frame_resource.y_stop]
                self.ingest_job.coord_frame_z_extent = [coord_frame_resource.z_start,
                                                        coord_frame_resource.z_stop]
                self.ingest_job.x_extent = self.ingest_job.coord_frame_x_extent
                self.ingest_job.y_extent = self.ingest_job.coord_frame_y_extent
                self.ingest_job.z_extent = self.ingest_job.coord_frame_z_extent

        return coord_frame_resource

    def setup_boss_experiment(self, get_only=True):
        # if we don't know the coordinate frame parameters, get the one with the same name
        if get_only:
            exp_setup = ExperimentResource(
                self.ingest_job.exp_name, self.ingest_job.coll_name, self.coord_frame_name)
        else:
            # if all elements are the same, isotropic, otherwise anisotropic
            if len(set(self.ingest_job.voxel_size)) <= 1:
                hierarchy_method = 'isotropic'
            else:
                hierarchy_method = 'anisotropic'

            num_hierarchy_levels = self.calc_hierarchy_levels()

            exp_setup = ExperimentResource(self.ingest_job.exp_name, self.ingest_job.coll_name, self.coord_frame_name, '',
                                           num_hierarchy_levels, hierarchy_method)
        exp_resource = self.get_boss_project(exp_setup, get_only)

        # record the offset (if there is any) into BOSS metadata field for experiment
        self.record_offsets(exp_resource)

        return exp_resource

    def record_offsets(self, exp_resource):
        if any([a > 0 for a in self.ingest_job.offsets]):
            offsets_dict = {'offsets': self.ingest_job.offsets}
            try:
                self.rmt.create_metadata(exp_resource, offsets_dict)
            except HTTPErrorList:  # keys already exist
                self.rmt.update_metadata(exp_resource, offsets_dict)

    def setup_boss_channel(self, ch_description='', get_only=True):
        ch_args = [self.ingest_job.ch_name,
                   self.ingest_job.coll_name, self.ingest_job.exp_name]
        if not get_only:
            ch_args.extend((self.ingest_job.ch_type, ch_description,
                            0, self.ingest_job.boss_datatype, self.ingest_job.res))

            # annotation data
            if self.ingest_job.source_channel is not None:
                ch_args.append([self.ingest_job.source_channel])

                # checking to make sure source channel exists (creating if needed)
                try:
                    source_args = [self.ingest_job.source_channel,
                                   self.ingest_job.coll_name, self.ingest_job.exp_name]
                    source_setup = ChannelResource(*source_args)
                    self.get_boss_project(source_setup, True)
                except:
                    self.ingest_job.send_msg(
                        'Creating a dummy source channel for this annotation because none exists yet')
                    source_args.extend(('image', '', 0, 'uint8', 0))
                    source_setup = ChannelResource(*source_args)
                    self.get_boss_project(source_setup, False)

        ch_setup = ChannelResource(*ch_args)
        ch_resource = self.get_boss_project(ch_setup, get_only)

        if get_only:
            self.ingest_job.boss_datatype = ch_resource.datatype
            if self.ingest_job.datatype is None:
                self.ingest_job.datatype = ch_resource.datatype
            self.ingest_job.res = ch_resource.base_resolution

        return ch_resource

    def calc_hierarchy_levels(self, lowest_res=512):
        img_size = [self.coord_frame_resource.x_stop - self.coord_frame_resource.x_start,
                    self.coord_frame_resource.y_stop - self.coord_frame_resource.y_start,
                    self.coord_frame_resource.z_stop - self.coord_frame_resource.z_start]

        max_xy = max(img_size[0:1])
        # we add one because 0 is included in the number of downsampling levels
        num_levels = max(1, math.ceil(math.log(max_xy / lowest_res, 2)) + 1)
        return num_levels



"""
Python 3 program to download & upload data from NeuroData
"""


version = "1.0.2"


def check_version():
    """
    Tells you if you have an old version of ndex.
    """
    import requests
    r = requests.get('https://pypi.python.org/pypi/ndexchange/json').json()
    r = r['info']['version']
    if r != version:
        print("A newer version of ndex is available. " +
              "'pip install -U ndexchange' to update.")
    return r



import os

from setuptools import find_packages, setup

import ndex

here = os.path.abspath(os.path.dirname(__file__))

# get the dependencies and installs
with open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]

setup(name='ndexchange',
      version=ndex.version,
      description='Python 3 program to download data from NeuroData',
      url='https://github.com/neurodata/ndex',
      author='Benjamin Falk',
      author_email='falk.ben@jhu.edu',
      license='Apache 2.0',
      packages=find_packages(),
      install_requires=install_requires,
      keywords=[
          'brain',
          'microscopy',
          'neuroscience',
          'connectome',
          'connectomics',
          'spatial',
          'EM',
          'electron',
          'calcium',
          'database',
          'boss'
      ],
      zip_safe=False,
      entry_points={'console_scripts':
                    ['ndpull=ndex.ndpull.ndpull:main',
                     'ndpush=ndex.ndpush.ingest_large_vol:main'], },
      )
